from pathlib import Path
from datetime import datetime
from zipfile import ZipFile
import requests
import pandas as pd
from tqdm import tqdm
import edinet


class Storage():

    def __init__(self, root=""):
        _root = root if root else "data"
        self._default_raw_data = f"{_root}/raw"
        self._default_interim_data = f"{_root}/interim"
        self._default_processed_data = f"{_root}/processed"

    def download(self, directory="", kind="F", year="", force=False):
        DOWNLAD_URL = "https://s3-ap-northeast-1.amazonaws.com/chakki.esg.financial.jp/dataset/release/chakki_esg_financial{}_{}.zip"  # noqa

        _kind = kind if len(kind) == 1 else kind[-1]
        default = self._default_interim_data if _kind == "E"\
                  else self._default_raw_data  # noqa
        dirname = directory if directory else default
        _directory = Path(dirname)
        if not _directory.is_absolute():
            _directory = Path.cwd().joinpath(dirname)

        year = str(year if year else datetime.now().year - 1)
        if _kind == "F":
            url = DOWNLAD_URL.format("", year)
        elif _kind == "E":
            url = DOWNLAD_URL.format("_extracted", year)
        else:
            raise Exception(f"Specified kind {kind} does not appropriate.")

        if len(kind) == 2 and kind[0] == "X":
            name_suffix = "_extracted" if _kind == "E" else ""
            url = DOWNLAD_URL.format(name_suffix, "example")

        if not _directory.exists():
            _directory.mkdir(parents=True, exist_ok=True)
        file_name = url.split("/")[-1]
        path = _directory.joinpath(file_name)

        if path.exists() and not force:
            print(f"Download file {file_name} is already exist.")
            return path
        elif _directory.joinpath(year).exists() and not force:
            print(f"Expanded directory {year} is already exist.")
            return _directory.joinpath(year)

        resp = requests.get(url, stream=True)
        if not resp.ok:
            raise Exception("Can not get dataset from {}.".format(url))

        total_size = int(resp.headers.get("content-length", 0))
        print(f"Download {file_name}")
        with path.open(mode="wb") as f:
            chunk_size = 1024
            limit = total_size / chunk_size
            for data in tqdm(resp.iter_content(chunk_size=chunk_size),
                             total=limit, unit="B", unit_scale=True):
                f.write(data)

        print(f"Expand {file_name}")
        with ZipFile(path) as zip:
            for f in tqdm(iterable=zip.namelist(), total=len(zip.namelist())):
                zip.extract(member=f, path=_directory)

        path.unlink()

        return _directory.joinpath(year)

    def parse(self, aspect_dot_element,
              source_directory="", target_directory="",
              year="", edinet_code="", sec_code="", jcn="",
              normalized=True):

        if not source_directory:
            s_dir = Path.cwd().joinpath(self._default_raw_data)
        else:
            s_dir = Path(source_directory)
            if not s_dir.is_absolute():
                s_dir = Path.cwd().joinpath(source_directory)

        if not target_directory:
            t_dir = Path.cwd().joinpath(self._default_interim_data)
        else:
            t_dir = Path(target_directory)
            if not t_dir.is_absolute():
                t_dir = Path.cwd().joinpath(target_directory)

        if not s_dir.exists():
            raise Exception(f"Files does not exist at {s_dir}")

        years = [year] if year else []
        if len(years) == 0:
            for d in s_dir.iterdir():
                if d.is_dir():
                    years.append(d.name)

        aspect, element = aspect_dot_element.split(".")
        sec_code = str(sec_code)
        filters = {
            "edinet_code": edinet_code,
            "sec_code": sec_code + "0" if len(sec_code) == 4 else sec_code,
            "jcn": str(jcn)
            }
        for y in years:
            s_y_dir = s_dir.joinpath(y)
            documents = pd.read_csv(s_y_dir.joinpath("documents.csv"),
                                    sep="\t",  header=0)
            documents["sec_code"] = documents["sec_code"].astype(str)
            documents["jcn"] = documents["jcn"].astype(str)

            for k in filters:
                if filters[k]:
                    documents = documents[documents[k] == filters[k]]

            t_y_dir = t_dir.joinpath(y)
            t_y_dir.mkdir(parents=True, exist_ok=True)
            with t_y_dir.joinpath("documents.csv").open(
                    mode="w", encoding="utf-8") as f:
                documents.to_csv(f, header=True, index=False, sep="\t")

            t_y_dir = t_y_dir.joinpath("docs")
            t_y_dir.mkdir(exist_ok=True)
            total = len(documents)
            print(f"Fiscal year {y}")
            for i, doc_id in tqdm(iterable=documents["doc_id"].iteritems(),
                                  total=total):
                f = s_y_dir.joinpath(f"docs/{doc_id}.xbrl")
                try:
                    text = edinet.parse(f.absolute(), aspect, element)
                except Exception as ex:
                    print(f"Failed to parse {f.name} of {aspect_dot_element}.")

                base_name = aspect_dot_element.replace(".", "_")
                if normalized:
                    path = t_y_dir.joinpath(f"{f.stem}_{base_name}.txt")
                    with path.open("w", encoding="utf-8") as f:
                        f.write(text.value)
                else:
                    path = t_y_dir.joinpath(f"{f.stem}_{base_name}.html")
                    with path.open("w", encoding="utf-8") as f:
                        f.write(text.ground)

        return t_dir

    def tokenize(self, tokenizer="janome",
                 source_directory="", target_directory="",
                 year="", edinet_code="", sec_code="", jcn="",
                 aspect_dot_element="",
                 mode="", dictionary="", dictionary_type=""):

        tokenize = None

        if tokenizer == "janome":
            from janome.tokenizer import Tokenizer

            if not dictionary:
                _tokenizer = Tokenizer(wakati=True)
            else:
                _tokenizer = Tokenizer(dictionary, udic_type=dictionary_type,
                                       udic_enc="utf8", wakati=True)

            def tokenize(text):
                return _tokenizer.tokenize(text)

        elif tokenizer == "sudachi":
            from sudachipy.tokenizer import Tokenizer
            from sudachipy.dictionary import Dictionary

            if mode == "A":
                mode = Tokenizer.SplitMode.A
            elif mode == "B":
                mode = Tokenizer.SplitMode.B
            elif mode == "C":
                mode = Tokenizer.SplitMode.C

            _tokenizer = Dictionary().create()

            def tokenize(text):
                return [m.surface() for m in
                        _tokenizer.tokenize(text, mode)]

        else:
            raise Exception(f"Tokenizer {tokenizer} does not supported.")

        if not source_directory:
            s_dir = Path.cwd().joinpath(self._default_interim_data)
        else:
            s_dir = Path(source_directory)
            if not s_dir.is_absolute():
                s_dir = Path.cwd().joinpath(source_directory)

        if not target_directory:
            t_dir = Path.cwd().joinpath(self._default_processed_data)
        else:
            t_dir = Path(target_directory)
            if not t_dir.is_absolute():
                t_dir = Path.cwd().joinpath(target_directory)

        if not s_dir.exists():
            raise Exception(f"Files does not exist at {s_dir}")

        years = [year] if year else []
        if len(years) == 0:
            for d in s_dir.iterdir():
                if d.is_dir():
                    years.append(d.name)

        aspect_element = aspect_dot_element.replace(".", "_")
        sec_code = str(sec_code)
        filters = {
            "edinet_code": edinet_code,
            "sec_code": sec_code + "0" if len(sec_code) == 4 else sec_code,
            "jcn": str(jcn)
            }
        suffix = "_" + aspect_element if aspect_element else "*"
        for y in years:
            s_y_dir = s_dir.joinpath(y)
            documents = pd.read_csv(s_y_dir.joinpath("documents.csv"),
                                    sep="\t",  header=0)
            documents["sec_code"] = documents["sec_code"].astype(str)
            documents["jcn"] = documents["jcn"].astype(str)
            for k in filters:
                if filters[k]:
                    documents = documents[documents[k] == filters[k]]

            t_y_dir = t_dir.joinpath(y)
            t_y_dir.mkdir(parents=True, exist_ok=True)
            with t_y_dir.joinpath("documents.csv").open(
                    mode="w", encoding="utf-8") as f:
                documents.to_csv(f, header=True, index=False, sep="\t")

            t_y_dir = t_y_dir.joinpath("docs")
            t_y_dir.mkdir(exist_ok=True)
            total = len(documents)
            print(f"Fiscal year {y}")
            for i, doc_id in tqdm(iterable=documents["doc_id"].iteritems(),
                                  total=total):
                for f in s_y_dir.joinpath("docs").glob(f"{doc_id}{suffix}.txt"):
                    tokens = []
                    with f.open(encoding="utf-8") as fobj:
                        lines = fobj.readlines()
                        lines = [ln.strip() for ln in lines]
                        lines = [ln for ln in lines if ln]
                        tokens = [tokenize(ln) for ln in lines]

                    path = t_y_dir.joinpath(f"{f.stem}_tokenized.txt")
                    with path.open("w", encoding="utf-8") as f:
                        for ts in tokens:
                            f.write("\t".join(ts) + "\n")

        return t_dir



import unittest
import shutil
from pathlib import Path
from chafic.storage import Storage


class TestStorage(unittest.TestCase):
    ROOT = "tdata"

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(Path.cwd().joinpath("tdata"))

    def test_download(self):
        path = self._download()
        self.assertTrue(path.exists())
        self.assertEqual(path.name, "2018")
        self.assertTrue(path.joinpath("documents.csv").exists())
        self.assertTrue(path.joinpath("docs").exists())
        self.assertGreater(
            len(list(path.joinpath("docs").glob("*.xbrl"))), 1)

    def test_download_extracted(self):
        storage = Storage(self.ROOT)
        path = storage.download(directory="tdata/rawe", kind="XE", year=2018)
        self.assertTrue(path.exists())
        self.assertEqual(path.name, "2018")
        self.assertTrue(path.joinpath("documents.csv").exists())
        self.assertTrue(path.joinpath("docs").exists())
        self.assertGreater(
            len(list(path.joinpath("docs").glob("*.txt"))), 1)

    def test_parse(self):
        storage = Storage(self.ROOT)
        path = self._download(kind="F")
        path = storage.parse("company.history")
        self.assertTrue(path.exists())
        self.assertTrue(path.joinpath("2018").exists())
        self.assertTrue(path.joinpath("2018/documents.csv").exists())
        self.assertGreater(
            len(list(path.joinpath("2018/docs").glob("*company_history.txt"))), 0)

        path = storage.parse("business.risks", sec_code="1376")
        with path.joinpath("2018/documents.csv").open(encoding="utf-8") as f:
            self.assertGreater(len(f.readlines()), 1)
        self.assertEquals(
            len(list(path.joinpath("2018/docs").glob("*business_risks.txt"))), 1)

    def test_tokenize(self):
        storage = Storage(self.ROOT)
        path = self._download(kind="F")
        path = storage.parse("company.history")
        path = storage.tokenize(tokenizer="janome")
        self.assertTrue(path.exists())
        self.assertTrue(path.joinpath("2018").exists())
        self.assertTrue(path.joinpath("2018/documents.csv").exists())
        self.assertGreater(
            len(list(path.joinpath("2018/docs").glob(
                "*company_history_tokenized.txt"))), 0)

        path = storage.parse("business.risks", sec_code="1376")
        path = storage.tokenize(tokenizer="sudachi")
        with path.joinpath("2018/documents.csv").open(encoding="utf-8") as f:
            self.assertGreater(len(f.readlines()), 1)
        self.assertEquals(
            len(list(path.joinpath("2018/docs").glob(
                "*business_risks_tokenized.txt"))), 1)

    def _download(self, kind="F"):
        storage = Storage(self.ROOT)
        path = storage.download(kind="X" + kind, year=2018)
        return path



import fire
from chafic.storage import Storage


class Chafic(object):
    """
    Data management tool for chaFiC dataset.
    """

    def __init__(self):
        self._storage = Storage()

    def download(self, directory="", kind="F", year="", force=False):
        """Download the {kind} {year} dataset to {directory}.

        Args:
            directory (str): Downloaded dataset to specified directory.
            kind (str): 'F': raw file datadata, 'E': text extracted data.
            year (str): Financial year of dataset.
            force (bool): When True, overwrite data if exist.

        Returns:
            str: Path to downloaded directory

        """
        return self._storage.download(directory, kind, year, force)

    def parse(self, aspect_element,
              source_directory="", target_directory="",
              year="", edinet_code="", sec_code="", jcn="",
              normalized=True):
        """
        Parse {aspect_to_element} from files in {source_directory}{year} and
        save it in {target_directory}{year} as txt/html file.

        Args:
            aspect_element (str): Target aspect.element (ex: company.history).
            source_directory (str): Source directory includes XBRL files.
            target_directory (str): Target directory that txt/htmls are saved.
            year (str): Target financial year.
            edinet_code (str): EDINET code to specify compan.
            sec_code (str): SEC code to specify compan.
            jcn (str): Target JCN code to specify compan.
            normalized: (bool): True: extract text, False: save raw xml(html).

        Returns:
            str: Path to parsed files directory

        """
        return self._storage.parse(
            aspect_element, source_directory, target_directory,
            year, edinet_code, sec_code, jcn,
            normalized)

    def tokenize(self, tokenizer="janome",
                 source_directory="", target_directory="",
                 year="", edinet_code="", sec_code="", jcn="",
                 aspect_element="",
                 mode="", dictionary="", dictionary_type=""):
        """
        Tokenize by {tokenizer} from files in {source_directory}{year} and
        save it in {target_directory}{year} as txt/html file.

        Args:
            tokenizer (str): Japanese tokenizer ('janome' or 'sudachi').
            source_directory (str): Source directory includes XBRL files.
            target_directory (str): Target directory that txt/htmls are saved.
            year (str): Target financial year.
            edinet_code (str): EDINET code to specify compan.
            sec_code (str): SEC code to specify compan.
            jcn (str): Target JCN code to specify compan.
            aspect_element (str): Target aspect.element (ex: company.history).
            mode: (str): Sudachi tokenizer mode.
            dictionary: (str): Dictionary path for Janome.
            dictionary_type: (str): Dictionary type for Janome.

        Returns:
            str: Path to parsed files directory

        """
        return self._storage.tokenize(
            tokenizer,
            source_directory, target_directory,
            year, edinet_code, sec_code, jcn, aspect_element,
            mode, dictionary, dictionary_type)


def main():
    fire.Fire(Chafic)


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages


readme = ""
with open("README.md", encoding="utf-8") as f:
    readme = f.read()


setup(
    name="chafic",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    packages=find_packages(exclude=["texts"]),
    description="chakki Financial Report Corpus",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="icoxfog417",
    author_email="icoxfog417@yahoo.co.jp",
    license="MIT",
    url="https://github.com/chakki-works/chaFiC",
    install_requires=[
        "edinet-python>=0.1.14",
        "fire>=0.2.1",
        "pandas>=0.25.1",
        "tqdm>=4.36.1"
    ],
    entry_points={"console_scripts": "chafic = chafic.cli:main"}
)
